@@ -335,11 +335,12 @@ replace_word(GtkWidget *menuitem, GtkSpell *spell) {
 	g_free(oldword);
 }
 
-static GtkWidget*
-build_suggestion_menu(GtkSpell *spell, GtkTextBuffer *buffer,
-                      const char *word) {
+/* This function populates suggestions at the top of the passed menu */
+static void
+add_suggestion_menus(GtkSpell *spell, GtkTextBuffer *buffer,
+                      const char *word, GtkWidget *topmenu) {
 	const char *suggestion;
-	GtkWidget *topmenu, *menu;
+	GtkWidget *menu;
 	GtkWidget *mi;
 	GtkWidget *hbox;
 	void *spelldata;
@@ -347,10 +348,12 @@ build_suggestion_menu(GtkSpell *spell, GtkTextBuffer *buffer,
 	size_t n_suggs, i;
 	char *label;
 	
-	topmenu = menu = gtk_menu_new();
+	menu = topmenu;
 
 	if (!spell->speller)
-		return topmenu;
+		return;
+
+	gint menu_position = 0;
 
 	suggestions = enchant_dict_suggest(spell->speller, word, strlen(word), &n_suggs);
 
@@ -363,18 +366,16 @@ build_suggestion_menu(GtkSpell *spell, GtkTextBuffer *buffer,
 		mi = gtk_menu_item_new();
 		gtk_container_add(GTK_CONTAINER(mi), label);
 		gtk_widget_show_all(mi);
-		gtk_menu_shell_prepend(GTK_MENU_SHELL(menu), mi);
+		gtk_menu_shell_insert(GTK_MENU_SHELL(menu), mi, menu_position++);
 	} else {
 		/* build a set of menus with suggestions. */
+		gboolean inside_more_submenu = FALSE;
 		for (i = 0; i < n_suggs; i++ ) {
 			if (i > 0 && i % 10 == 0) {
-				mi = gtk_menu_item_new();
-				gtk_widget_show(mi);
-				gtk_menu_shell_append(GTK_MENU_SHELL(menu), mi);
-
+				inside_more_submenu = TRUE;
 				mi = gtk_menu_item_new_with_label(_("More..."));
 				gtk_widget_show(mi);
-				gtk_menu_shell_append(GTK_MENU_SHELL(menu), mi);
+				gtk_menu_shell_insert(GTK_MENU_SHELL(menu), mi, menu_position++);
 
 				menu = gtk_menu_new();
 				gtk_menu_item_set_submenu(GTK_MENU_ITEM(mi), menu);
@@ -383,17 +384,13 @@ build_suggestion_menu(GtkSpell *spell, GtkTextBuffer *buffer,
 			g_signal_connect(G_OBJECT(mi), "activate",
 					G_CALLBACK(replace_word), spell);
 			gtk_widget_show(mi);
-			gtk_menu_shell_append(GTK_MENU_SHELL(menu), mi);
+			if (inside_more_submenu) gtk_menu_shell_append(GTK_MENU_SHELL(menu), mi);
+			else gtk_menu_shell_insert(GTK_MENU_SHELL(menu), mi, menu_position++);
 		}
 	}
 
 	enchant_dict_free_suggestions(spell->speller, suggestions);
 
-	/* Separator */
-	mi = gtk_menu_item_new();
-	gtk_widget_show(mi);
-	gtk_menu_shell_append(GTK_MENU_SHELL(topmenu), mi);
-
 	/* + Add to Dictionary */
 	label = g_strdup_printf(_("Add \"%s\" to Dictionary"), word);
 	mi = gtk_image_menu_item_new_with_label(label);
@@ -403,7 +400,7 @@ build_suggestion_menu(GtkSpell *spell, GtkTextBuffer *buffer,
 	g_signal_connect(G_OBJECT(mi), "activate",
 			G_CALLBACK(add_to_dictionary), spell);
 	gtk_widget_show_all(mi);
-	gtk_menu_shell_append(GTK_MENU_SHELL(topmenu), mi);
+	gtk_menu_shell_insert(GTK_MENU_SHELL(topmenu), mi, menu_position++);
 
 	/* - Ignore All */
 	mi = gtk_image_menu_item_new_with_label(_("Ignore All"));
@@ -412,18 +409,83 @@ build_suggestion_menu(GtkSpell *spell, GtkTextBuffer *buffer,
 	g_signal_connect(G_OBJECT(mi), "activate",
 			G_CALLBACK(ignore_all), spell);
 	gtk_widget_show_all(mi);
-	gtk_menu_shell_append(GTK_MENU_SHELL(topmenu), mi);
+	gtk_menu_shell_insert(GTK_MENU_SHELL(topmenu), mi, menu_position++);
+}
+
+static GtkWidget*
+build_suggestion_menu(GtkSpell *spell, GtkTextBuffer *buffer,
+                      const char *word) {
+	GtkWidget *topmenu;
+	topmenu = gtk_menu_new();
+	add_suggestion_menus(spell, buffer, word, topmenu);
 
 	return topmenu;
 }
 
+static void
+language_change_callback(GtkWidget *mi, GtkSpell* spell) {
+	GError* error = NULL;
+	gchar *name;
+	g_object_get(G_OBJECT(mi), "name", &name, NULL);
+	gtkspell_set_language(spell, name, &error);
+	g_free(name);
+}
+
+struct _languages_cb_struct {GtkWidget *menu; GtkSpell *spell;};
+
+static void
+dict_describe_cb(const char * const lang_tag,
+		 const char * const provider_name,
+		 const char * const provider_desc,
+		 const char * const provider_file,
+		 void * user_data) {
+
+	GtkWidget* mi;
+	struct _languages_cb_struct *languages_cb_struct = (struct _languages_cb_struct *)user_data;
+	GtkWidget* menu = languages_cb_struct->menu;
+	GtkSpell* spell = languages_cb_struct->spell;
+
+	/* TODO get the language code translation to the current locale */
+	/* TODO add a radiobutton with the currently selected language */
+	mi = gtk_menu_item_new_with_label(lang_tag);
+	g_object_set(G_OBJECT(mi), "name", lang_tag, NULL);
+	g_signal_connect(G_OBJECT(mi), "activate",
+		G_CALLBACK(language_change_callback), spell);
+	gtk_widget_show(mi);
+	gtk_menu_shell_append(GTK_MENU_SHELL(menu), mi);
+}
+
+static GtkWidget*
+build_languages_menu(GtkSpell *spell) {
+	GtkWidget* menu = gtk_menu_new();
+
+	struct _languages_cb_struct languages_cb_struct;
+	languages_cb_struct.menu = menu;
+	languages_cb_struct.spell = spell;
+
+	enchant_broker_list_dicts(spell->broker, dict_describe_cb, &languages_cb_struct);
+
+	return menu;
+}
+
 static void
 populate_popup(GtkTextView *textview, GtkMenu *menu, GtkSpell *spell) {
-	GtkWidget *img, *mi;
+	GtkWidget *mi;
 	GtkTextBuffer *buffer = gtk_text_view_get_buffer(textview);
 	GtkTextIter start, end;
 	char *word;
 
+	/* menu separator comes first. */
+	mi = gtk_separator_menu_item_new();
+	gtk_widget_show(mi);
+	gtk_menu_shell_prepend(GTK_MENU_SHELL(menu), mi);
+
+	/* on top: language selection */
+	mi = gtk_menu_item_new_with_label(_("Languages"));
+	gtk_menu_item_set_submenu(GTK_MENU_ITEM(mi), build_languages_menu(spell));
+	gtk_widget_show_all(mi);
+	gtk_menu_shell_prepend(GTK_MENU_SHELL(menu), mi);
+
 	/* we need to figure out if they picked a misspelled word. */
 	get_word_extents_from_mark(buffer, &start, &end, spell->mark_click);
 
@@ -432,23 +494,10 @@ populate_popup(GtkTextView *textview, GtkMenu *menu, GtkSpell *spell) {
 	if (!gtk_text_iter_has_tag(&start, spell->tag_highlight))
 		return; /* word wasn't misspelled. */
 
-	/* menu separator comes first. */
-	mi = gtk_menu_item_new();
-	gtk_widget_show(mi);
-	gtk_menu_shell_prepend(GTK_MENU_SHELL(menu), mi);
-
-	/* then, on top of it, the suggestions menu. */
-	img = gtk_image_new_from_stock(GTK_STOCK_SPELL_CHECK, GTK_ICON_SIZE_MENU);
-	mi = gtk_image_menu_item_new_with_label(_("Spelling Suggestions"));
-	gtk_image_menu_item_set_image(GTK_IMAGE_MENU_ITEM(mi), img);
-
+	/* then, on top of it, the suggestions */
 	word = gtk_text_buffer_get_text(buffer, &start, &end, FALSE);
-	gtk_menu_item_set_submenu(GTK_MENU_ITEM(mi),
-			build_suggestion_menu(spell, buffer, word));
+	add_suggestion_menus(spell, buffer, word, GTK_WIDGET (menu) );
 	g_free(word);
-
-	gtk_widget_show_all(mi);
-	gtk_menu_shell_prepend(GTK_MENU_SHELL(menu), mi);
 }
 
 /* when the user right-clicks on a word, they want to check that word.
